## Copyright (c) 2020
## Global Climate Forum e.V.
## http://www.globalclimateforum.org

## Authors:
##     Steffen Fuerst

## This file is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This file is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## version 0.11

if (!require("tidyverse")) install.packages("tidyverse")
if (!require("magrittr")) install.packages("magrittr")
library(tidyverse)
library(magrittr)
options(crayon.enabled = FALSE)
options(tibble.width = Inf)

## The ~initSimulation~ function creates a new R environment, which
## contains the complete state of the simulation and some functions to
## access the three different layers. It also incl. the tables that
## allocate the nodes and edges of the *Interagent Layer* to the
## different (virtual) PEs.
initSimulation <- function(agentTypes,
                           edgeTypes,
                           modelParams,
                           stateInitFunc,
                           globalTypes,
                           globalInitFunc = NULL,
                           rasterTypes = NULL,
                           rasterInitFunc = NULL) {
  sim <- new.env()
  sim$agentTypes <- agentTypes
  sim$edgeTypes <- edgeTypes
  sim$globalTypes <- globalTypes
  sim$rasterTypes <- rasterTypes
  sim$numPEs <- modelParams$numPEs

  modelStateFull <- stateInitFunc(modelParams)
  ## sim$distributed is a list of two vectors. The vectors are representing
  ## the different PEs, so the length of those vectors is equal to sim$numPEs.
  ## The first vector contains the model state of the Interagent Layer distributed 
  ## to the different PEs, the second contains for each vector element 
  ## (representing a PE) a table that contains for all edges where a node of 
  ## this PE is involved the information on which PE the node of the opposite 
  ## side is "living".
  sim$distributed <- .mapAgentsToPEs(modelStateFull, agentTypes, sim$numPEs)
  sim$getExplicitLayer <- function() { sim$distributed[[1]] }
  sim$getPEs <- function() { sim$distributed[[2]] }

  if (! is.null(globalInitFunc)) {
    sim$global <- globalInitFunc(modelParams)
  }
  ## we always create automatically the type MODELPARAMS in the Global Layer
  sim$global$MODELPARAMS <- modelParams

  if (! is.null(rasterTypes)) {
    sim$raster <- rasterInitFunc(modelParams)
    for (r in rasterTypes) {
      ## .createRasterLayer modifies the sim environment as a side effect
      ## by adding for each rasterType a table with nodes for this raster and
      ## a table with a PE list
      sim <- .createRasterLayer(sim, r, sim$raster[[r]], 1)
    }
  }

  sim$getGlobalLayer <- function() { sim$global }
  sim$getRasterLayer <- function() { sim$raster }
  
  ## reduce is a helper function to calculate global values based on the state
  ## of the nodes in the *Interagent Layer*.
  sim$reduce <- function(type, agentFunc, aggregationFunc) {
    sim$getExplicitLayer() %>% map(type) %>% map(agentFunc) %>% unlist %>% aggregationFunc
  }
  
  sim
}


## The helper function ~.mapAgentsToPEs~ takes a
## non-distributed model state and creates a distributed version. 
## See also the comment in initSimulation.
.mapAgentsToPEs <- function(modelState, agentTypes, numPEs) {
  edgeTypes <- modelState %>% names %>% discard(. %in% agentTypes)

  ## Store which agent was moved to which PE
  PElist <- tibble(ID = character(), 
                   PE = integer())

  ## Prepare the distributed data structures (the indices are the different PEs)
  modelStateDistributed <- vector("list", numPEs)
  PElists <- vector("list", numPEs)
  for (i in 1:numPEs) {
    PElists[[i]] <- tibble(ID = character(), 
                           PE = integer())
  }

  ## Distribute the agents
  for (a in agentTypes) {
    ## We group the agents by their rowid (modulo numPEs), so that each PE gets the 
    ## same number (plus/minus one) of agents
    modelState[[a]] %<>% rowid_to_column("PE") %>%
      mutate(PE = ((PE - 1) %% numPEs) + 1) 
    PElist %<>% bind_rows(modelState[[a]]) %>% select(ID, PE)
    perPE <- modelState[[a]] %>% split(.$PE)

    for (i in 1:numPEs) {
      modelStateDistributed[[i]][[a]] <- perPE[[i]] 
      PElists[[i]] %<>% bind_rows(modelStateDistributed[[i]][[a]]) %>%
        select(ID, PE)
    }
  }

  ## Distribute the edges
  for (e in edgeTypes) {
    if (nrow(modelState[[e]]) > 0) {
      for (i in 1:numPEs) {
        ## Find for each PE the edges where an agent of this PE is involved ..
        fID <- modelState[[e]] %>% filter(fromID %in% PElists[[i]]$ID)
        tID <- modelState[[e]] %>% filter(toID %in% PElists[[i]]$ID)
        ## ... and add those edges to the distributed state
        modelStateDistributed[[i]][[e]] <- bind_rows(fID, tID) %>% unique
        ## Also add the PEs of the agents on the other edge pos to the PElists
        PElists[[i]] %<>% bind_rows(fID %>% transmute(ID = toID) %>% left_join(PElist, by = "ID")) %>% unique
        PElists[[i]] %<>% bind_rows(tID %>% transmute(ID = fromID) %>% left_join(PElist, by = "ID")) %>% unique
      }
    }
  }

  list(modelStateDistributed, PElists)
}


## create a unique name for the single raster nodes 
.createRasterID <- function(raster, x, y) {
  paste0(raster,"x", x,"y", y)
}

## create for a raster a tibble that contains the distribution
## of the nodes to PEs. In the current implementation, we distribute
## all nodes to PE 1, as we don't have a real parallel implemention anyway
.createRasterLayer <- function(sim, rname, rdata, distance) {
  ## create a unique name for the table of PEs for the nodes of a raster
  PEsName <- paste0(rname, "PEs")

  ## Iterate over all elements of the raster and add the 
  ## PE information to the sim environment
  for (y in seq(dim(rdata)[1])) {
    for (x in seq(dim(rdata)[2])) {
      id = .createRasterID(rname, x, y)
      sim[[PEsName]] %<>% bind_rows(tibble(ID = id, PE = 1))
    }
  }
  sim
}

## a platform helper function for the initialization, that takes a non-distributed
## model state where the nodes does not have an ID, and add the node tables
## a column "ID", which contains unique IDs for each node
withStringID <- function(modelState) {
  modelState %>%
    imap(~ rowid_to_column(.x, var = "ID") %>% mutate(ID = paste0(.y,ID)))
}


## To modify the state of the agents and edges in the *Interagent Layer*
## we call the platform function ~applyTransition~:
applyTransition <- function(sim, 
                            transitionFunction,
                            networks = NULL,
                            invariantTypes = NULL,
                            distance = 1) {
  ## ensure that the edges are known 
  sim$distributed <- .distributeNetworkState(sim)

  ## in the case that the network is a raster, we must add the
  ## PEs of the nodes of this raster to the known PEs table 
  for (n in networks) {
    PEsName <- paste0(n, "PEs")
    if (! is.null(sim[[PEsName]])) {
      for (i in 1:sim$numPEs) {
        sim$distributed[[2]][[i]] %<>% bind_rows(sim[[PEsName]])
      }
    }
  }

  modelStateDistributed <- sim$getExplicitLayer()


  ## The .enhanceAgentStates function add that the state of agents
  ## at the tail position of one network to the corresponding PE.
  if (! is.null(networks)) {
    enhancedModelStateD <- .enhanceAgentStates(sim, networks)
  } else {
    enhancedModelStateD <- sim$getExplicitLayer()
  }

  ## Now we simulate the parallel processing of the transition
  ## function by iterating of the different distributed model state
  for (i in 1:sim$numPEs) {
    ## we first store the state of the agents or edges which 
    ## are invariant in this transition
    invariantState <- invariantTypes %>%
        map(~ modelStateDistributed[[i]][[.x]]) %>%
        set_names(invariantTypes)

    ## call all agents to apply their transitionFunction
    modelStateDistributed[[i]] <- sim$agentTypes %>%
      map(~ .transform(sim,
                       .x,
                       modelStateDistributed[[i]][[.x]],
                       enhancedModelStateD[[i]],
                       transitionFunction,
                       networks,
                       distance)) %>%
      .consolidate(sim)
    
    ## for the invariant state, add this to the state constructed 
    ## by the agents (in the case, that the agents also returned
    ## entities of an invariantType, those will be overwritten)
    for (n in invariantTypes)
        modelStateDistributed[[i]][[n]] <- invariantState[[n]]
  }
  sim$distributed[[1]] <- modelStateDistributed

  ## allow method chaining by returning the sim environment
  sim
}

## ~.distributeNetworkState~ takes the result of a transition
## function, and sends the edges to the PEs of the agents on the head
## or tail position.
.distributeNetworkState <- function(sim) {
  modelStateDistributed <- sim$getExplicitLayer()
  PElists <- sim$getPEs()
  modelStateEnhanced <- modelStateDistributed
  ## create new empty lists
  newPElists <- PElists %>% map(~ .x[0,])

  for (i in 1:sim$numPEs) {
    for (e in sim$edgeTypes) {
      ## We allow the creation of Networks while the simulation is running
      if (! is.null(modelStateDistributed[[i]][[e]]) &&
          nrow(modelStateDistributed[[i]][[e]]) > 0) {
        ## Create two version of the edge tibble, ~matchFrom~ is extended with 
        ## the PEs of the tail, ~matchTo~ with the PEs of the head agents
        matchFrom <- inner_join(modelStateDistributed[[i]][[e]],
                                PElists[[i]], by = c("fromID" = "ID"))
        matchTo <- inner_join(modelStateDistributed[[i]][[e]],
                              PElists[[i]], by = c("toID" = "ID"))
        ## sendNetworkTable is a tibble: fromID toID PE
        sendNetworkTable <- bind_rows(matchFrom, matchTo)
        ## sendPElist is a tibble: ID, PE, toPE
        sendPElist <- bind_rows(sendNetworkTable %>% select(ID = fromID,
                                                            toPE = PE),
                                sendNetworkTable %>% select(ID = toID,
                                                            toPE = PE)) %>%
          inner_join(PElists[[i]], by = "ID") 

        ## Use the send... tibbles to distribute the edges to the PEs
        ## and also to send the (agentID, PE) information to those PEs
        for (j in 1:sim$numPEs) {
          if (i != j) {
            modelStateEnhanced[[j]][[e]] %<>%
              bind_rows(sendNetworkTable %>% filter(PE == j) %>% select(-PE)) %>%
              unique
          }
          newPElists[[j]] %<>%
            bind_rows(sendPElist %>% filter(toPE == j) %>% select(-toPE)) %>%
            unique
        }
      }
    }
  }

  list(modelStateEnhanced, newPElists)
}

## The ~.enhanceAgentStates~ function ensures that the state of
## agents at the tail position of the network associated with the next
## transition function is available on the PEs of the agents at the
## head position of the edges. This function is called before the
## transition function itself, which then gets the returned extendend
## model state.
.enhanceAgentStates <- function(sim, edgeTypes) {
  modelStateDistributed <- sim$getExplicitLayer()
  PElists <- sim$getPEs()  
  modelStateEnhanced <- modelStateDistributed
  for (e in edgeTypes) {
    for (i in 1:sim$numPEs) {
      ## We allow the creation of Networks while the simulation is running
      if (! is.null(modelStateDistributed[[i]][[e]]) &&
          nrow(modelStateDistributed[[i]][[e]]) > 0) {
        ## agentPEmap is a tibble(fromID, toID, PE of toID agent)  
        agentPEmap <- modelStateDistributed[[i]][[e]] %>%
          select(fromID, toID) %>%
          left_join(PElists[[i]], by = c("toID" = "ID")) %>%
          unique

        for (a in sim$agentTypes) {
          ## iterate over the other PEs
          for (j in (1:sim$numPEs %>% discard(. == i))) {
            ## find the agents that must be send to other PEs ...
            agentsIDtoSend <- agentPEmap %>% filter(PE == j) %>% select(fromID)
            agentsToSend <- agentsIDtoSend %>% 
              inner_join(modelStateDistributed[[i]][[a]], by = c("fromID" = "ID")) %>%
              unique %>%
              rename(ID = fromID)
            ## ... and add them to the state of this other PE  
            modelStateEnhanced[[j]][[a]] %<>% bind_rows(agentsToSend)
          }
        }
      }
    }
  }

  modelStateEnhanced
}

## ~.transform~ takes one agent type and iterates over the agents of
## this type, calling for each agent one of the transition
## functions. Inside of ~.transform~ we use the ~limitState~ function
## to ensure, that only the state of agents in the tail position of
## the associated network is accessable.
.transform <- function(sim,
                       type,
                       agents,
                       modelState,
                       transitionFunc,
                       networks,
                       distance = 1) {
  ## create the raster nodes that are embedded in the part of the model 
  ## state that is visible for the agent
  createLocationNodes <- function(agent) {
    rasterTables <- NULL
    for (n in networks) {
      if (n %in% sim$rasterTypes) {
        ## in the case that the network is a raster, we must create
        ## the explicit network first
        rasterTables[[n]] <- NULL
        r <- sim$raster[[n]]

        x <- agent[[paste0(n,"x")]]
        ## it's optional to assign agents to a raster, so only
        ## continue, if this is really the case
        if (! is.null(x)) {
          y <- agent[[paste0(n,"y")]]
          ## calculate the area of the Moore neighborhood (with borders)
          fromx <- max(x - distance, 1)
          tox <- min(x + distance, dim(r)[2])
          fromy <- max(y - distance, 1)
          toy <- min(y + distance, dim(r)[1])

          ## iterate over this area and create nodes the nodes
          for (x in seq(fromx, tox)) {
            for (y in seq(fromy, toy)) {
              id = .createRasterID(n, x, y)
              rasterTables[[n]] %<>% bind_rows(tibble(ID = id, 
                                                      x = x,
                                                      y = y,
                                                      value = r[[y, x]]))
            }
          }
        }
      }
    }
    rasterTables
  }

  ## ensure, that only the state of agents in the tail position of the
  ## associated network is accessable.
  limitState <- function(agent) {
    agentTables <- NULL
    if (! is.null(networks)) {
      for (n in networks) {
        ## a network name can be also a raster, but the raster nodes 
        ## are allready handled in createLocationNodes above
        if (!(n %in% sim$rasterTypes)) {
          incoming <- modelState[[n]] %>% filter(toID == agent$ID)

          if (incoming %>% length > 0) # check that we have any edge
            ## filter the agents, so that only agents which are 
            ## in the ~incoming~ tibble are part of the ~agentTables~
            agentTables <- sim$agentTypes %>%
              map(~ bind_rows(agentTables[[.x]],
                              modelState[[.x]] %>% filter(ID %in% incoming$fromID))) %>%
              setNames(sim$agentTypes)
        }
      }
    }

    ## filter the network to edges where the agent is on tail or head position
    availableEdgeTypes <- sim$edgeTypes %>% keep(~ .x %in% networks)
    edgeTables <- availableEdgeTypes %>% 
      map(~ modelState[[.x]] %>% 
            (function(x) { if (is.null(x) || nrow(x) == 0) NULL else
                                                                { x %>% filter(agent$ID == fromID | 
                                                                               agent$ID == toID)}})) %>%
      setNames(availableEdgeTypes)

    ## create the overall available modelstate for the agent by concatenating
    ## the filtered nodes and edges from the Interagent Layer, the created nodes
    ## from the Spatial Layer and all nodes from the Global Layer
    c(agentTables, edgeTables, createLocationNodes(agent), sim$global)
  }

  ## iterate over all agents and call their transition function with the part
  ## of the model state, that is accessable by the agent
  agents %>%
    transpose %>%
    map(~ transitionFunc(as_tibble(.), type, limitState(.x))) 
}

## ~.consolidate~ gets the collected output of ~.transform~ from all
## agents in the system, so ~newModelState~ is a list (with the agent
## types as keys) of lists (with the agents as keys) of tables (the
## tables returned by a transition function), and reduces this to a
## single table per type.
.consolidate <- function(newModelState, sim) { 
  flattenLists <- newModelState %>% flatten

  c(sim$agentTypes, sim$edgeTypes) %>% 
    map(~ { flattenLists %>% map(.x) %>% bind_rows }) %>%
    setNames(c(sim$agentTypes, sim$edgeTypes))
}


applyRasterTransition <- function(sim, raster, transitionFunction,
                                  agentTypes = NULL,
                                  networks = NULL) {
  el <- sim$getExplicitLayer()
  ## generate the column names for the raster position in the 
  ## agent state
  xName <- paste0(raster, "x")
  yName <- paste0(raster, "y")

  ## update the value of the node at the x,y position as a side effect and 
  ## return the edges constructed by the transition function
  updatePos <- function(x, y) {
    ## create a temporary node for this raster position
    nodeName <- .createRasterID(raster, x, y)
    node <- tibble(ID = nodeName, x = x, y = y, value = sim$raster[[raster]][[y, x]])

    ## gather all the agent of the given agentTypes that are located 
    ## on this position
    agentTables <- NULL
    for (t in agentTypes) {
      for (i in 1:sim$numPEs) {
        agentTables[[t]] %<>% 
          bind_rows(el[[i]][[t]] %>% filter(get(xName) == !!x, get(yName) == !!y))
      }
    }

    ## gather all edges of the given networks which have the raster node on 
    ## the head or tail position
    edgeTables <- NULL
    for (t in networks) {
      for (i in 1:sim$numPEs) {
        if (! is.null(el[[i]][[t]])) {
          fromNode <- el[[i]][[t]] %>% filter(fromID == !!nodeName)
          toNode <- el[[i]][[t]] %>% filter(toID == !!nodeName)
          edgeTables[[t]] %<>% bind_rows(fromNode)
          edgeTables[[t]] %<>% bind_rows(toNode)
          ## in the case, that the raster node is on the head position, we
          ## must also add the PE of the node on the tail position to the 
          ## list of known PEs
          if (nrow(toNode) > 0)
            sim$distributed[[2]][[1]] %<>% 
              bind_rows(toNode %>% transmute(ID = fromID, PE = i)) %>% unique
        }
      }
    }

    ## call the transition function with the gathers agents, edges and the global layer
    r <- transitionFunction(node, c(agentTables, edgeTables, sim$global))
    ## update the value
    sim$raster[[raster]][[y, x]] <- r[["VALUE"]]
    ## and remove the value from the list, so that we return only the edges
    r[["VALUE"]] <- NULL
    r
  }

  ## the updatePos function can return edges, we gather them in the 
  ## combined list
  combined <- list()
  ## iterate over all positions
  for (y in seq(dim(sim$raster[[raster]])[1])) {
    for (x in seq(dim(sim$raster[[raster]])[2])) {
      ## updatePos can return lists with edges of arbitrary networks
      network <- updatePos(x, y)
      ## check for which networks edges was in the returned list
      networkTypes <- names(network)
      ## and add those to the combined 
      combined <- networkTypes %>% 
        map(~ bind_rows(combined[[.x]], network[[.x]])) %>% 
        setNames(networkTypes)
    }
  }
  ## clear all networks for which we got at least on edge returned 
  ## by the transition function
  for (n in names(combined))
    sim$distributed[[1]][[1]][[n]] <- NULL
  ## and add the new networks instead 
  sim$distributed[[1]][[1]] %<>% append(combined)

  ## allow method chaining by returning the sim environment
  sim
}

updateGlobal <- function(sim, type, func) {
  sim$global[[type]] <- func(sim$global[[type]], sim)
  sim
}
